@@ -1,7 +1,8 @@
 import os
 import logging
-import enum
+import json
 from inspect import getmembers, isclass
+import asyncio
 
 from pydantic import BaseSettings
 
@@ -10,26 +11,30 @@ from elasticsearch.exceptions import NotFoundError
 
 from fastapi import FastAPI, Depends, status
 from fastapi.security import HTTPBearer, OAuth2PasswordBearer
-from fastapi.security.http import HTTPAuthorizationCredentials
 from fastapi.encoders import jsonable_encoder
 from fastapi.middleware.cors import CORSMiddleware
 
 from bdating_common.model import HealthResponse, ConsumerProfile, MultiLangEnum
 from bdating_common import model as model_module
-#from bdating_common import model_zh as model_zh_module
-#from bdating_common import model_en as model_en_module
+# from bdating_common import model_zh as model_zh_module
+# from bdating_common import model_en as model_en_module
 
 from bdating_common.auth0_token_helper import IllegalTokenExcpetion, Auth0TokenVerifier
 from fastapi.responses import JSONResponse
 from bdating_common.es_helper import es_get_result_to_dict
-
+from fastapi import WebSocket
+from fastapi.websockets import WebSocket, WebSocketDisconnect
 import redis
+import aioredis
+from aioredis.pubsub import Channel
 from auth0.v3.authentication import Users as Auth0Users
+from typing import Union, Any, Dict
 
 oauth2_scheme = OAuth2PasswordBearer(tokenUrl='token')
 
 log = logging.getLogger(__name__)
 
+
 def get_model_enum_values():
     model_enums = {}
     for name, member in getmembers(model_module, isclass):
@@ -53,10 +58,12 @@ def get_model_enum_values():
                 log.warning('Enum %s has no proper translation defined', name)
                 res = []
                 for m in member.__members__:
-                    res.append({'_key': m, 'en': m.replace('_', ' ').capitalize(), 'zh': m.replace('_', ' ').capitalize()})
+                    res.append({'_key': m, 'en': m.replace(
+                        '_', ' ').capitalize(), 'zh': m.replace('_', ' ').capitalize()})
                 model_enums[name] = res
     return model_enums
 
+
 """
 def get_model_enum_values2():
     model_enums = {}
@@ -73,6 +80,7 @@ def get_model_enum_values2():
     return model_enums
 """
 
+
 class Settings(BaseSettings):
     app_name: str = 'StandardBdatingAPI'
     app_namespace: str = os.getenv('NAMESPACE', "")
@@ -94,7 +102,8 @@ settings = Settings()
 token_auth_scheme = HTTPBearer()
 token_verifier = Auth0TokenVerifier()
 
-token_cache_redis = None # TODO rework the global clients.
+token_cache_redis = None  # TODO rework the global clients.
+
 
 def _get_uid(token: str):
     global token_cache_redis
@@ -107,22 +116,55 @@ def _get_uid(token: str):
                 )
         cached_value = token_cache_redis.get(token)
         if cached_value is not None:
-            return cached_value.decode('utf-8')
+            return json.loads(cached_value.decode('utf-8')).get(f"{settings.auth0_namespace}uid")
     try:
         payload = token_verifier.verify(token)
-        print(payload)
         if payload.get("status") == 'error':
             raise IllegalTokenExcpetion()
         uid = payload.get(f"{settings.auth0_namespace}uid")
         log.warn('uid retrieved is %s', uid)
         if token_cache_redis is not None and uid:
-            ttl = 86400 * 1000 # hard code 1 day, ttl value is in milliseconds
-            token_cache_redis.setex(token, ttl, uid)
+            ttl = 86400 * 1000  # hard code 1 day, ttl value is in milliseconds
+            token_cache_redis.setex(
+                token, ttl, json.dumps(payload, default=str))
         return uid
     except Exception as e:
         log.warn('Token verification failed', exc_info=True)
         raise IllegalTokenExcpetion(e)
 
+
+async def validate_socket_token(token: str, websocket: WebSocket):
+    global token_cache_redis
+    if settings.redis_host:
+        if token_cache_redis is None:
+            token_cache_redis = redis.Redis(
+                settings.redis_host, settings.redis_port,
+                db=settings.token_cache_db_id,
+                password=settings.redis_password
+                )
+        cached_value = token_cache_redis.get(token)
+        if cached_value is not None:
+            payload = json.loads(cached_value.decode('utf-8'))
+            return payload.get(f"{settings.auth0_namespace}wallet", payload.get('wallet'))
+    try:
+        payload = token_verifier.verify(token)
+        if payload.get("status") == 'error':
+            raise IllegalTokenExcpetion(payload)
+        wallet = payload.get(
+            f"{settings.auth0_namespace}wallet", payload.get('wallet'))
+        log.warn('wallet retrieved is %s', wallet)
+        if token_cache_redis is not None and wallet:
+            ttl = 86400 * 1000  # hard code 1 day, ttl value is in milliseconds
+            token_cache_redis.setex(
+                token, ttl,  json.dumps(payload, default=str))
+        return wallet
+    except Exception as ex:
+        log.exception(ex)
+        await websocket.send_text('{"status":"403"}')
+        await websocket.close()
+        raise IllegalTokenExcpetion(ex)
+
+
 def get_uid(token: str, token_verifier: object):
     global token_cache_redis
     if settings.redis_host:
@@ -134,7 +176,7 @@ def get_uid(token: str, token_verifier: object):
                 )
         cached_value = token_cache_redis.get(token.credentials)
         if cached_value is not None:
-            return cached_value.decode('utf-8')
+            return json.loads(cached_value.decode('utf-8')).get(f"{settings.auth0_namespace}uid")
     try:
         result = token_verifier.verify(token.credentials)
         log.warn('result %s', result)
@@ -146,16 +188,18 @@ def get_uid(token: str, token_verifier: object):
         userinfo = users.userinfo(token.credentials)
         log.warn('userinfo %s', userinfo);
         uid = userinfo.get(f"{settings.auth0_namespace}uid")
-        #value = result.get(f"{settings.auth0_namespace}uid")
-        #log.warn('uid retrieved is %s', value)
+        # value = result.get(f"{settings.auth0_namespace}uid")
+        # log.warn('uid retrieved is %s', value)
         if token_cache_redis is not None and uid:
-            ttl = 86400 * 1000 # hard code 1 day, ttl value is in milliseconds
-            token_cache_redis.setex(token.credentials, ttl, uid)
+            ttl = 86400 * 1000  # hard code 1 day, ttl value is in milliseconds
+            token_cache_redis.setex(
+                token, ttl, json.dumps(payload, default=str))
         return uid
     except Exception as e:
         log.warn('Token verification failed', exc_info=True)
         raise IllegalTokenExcpetion(e)
 
+
 def get_contact(token: str):
     try:
         payload = token_verifier.verify(token)
@@ -167,6 +211,58 @@ def get_contact(token: str):
         log.warn('Token verification failed', exc_info=True)
         raise IllegalTokenExcpetion(e)
 
+
+async def ws_reader(channel: Channel, ws: WebSocket):
+    while await channel.wait_message():
+        text = None
+        try:
+            text = (await channel.get()).decode('utf-8')
+            msg = json.loads(text)
+            await ws.send_text(json.dumps({**msg, "channel": channel.name.decode('utf-8')}, default=str))
+        except Exception as e:
+            log.error(
+                'Could not process channel mesage from %s. Message: %s. Error: e', channel.name, text, e)
+
+
+async def redis_connector(
+    websocket: WebSocket, wallet: str
+):
+    async def consumer_handler(ws: WebSocket, redis_client: object):
+        try:
+            own_channel = (await redis_client.subscribe(wallet))[0]
+            asyncio.create_task(ws_reader(own_channel, ws))
+            additional_channel = None
+            while True:
+                message = await ws.receive_text()
+                message = message.strip()
+                if message.lower() == "close":
+                    await websocket.close()
+                    break
+                else:
+                    if additional_channel:
+                        redis_client.unsubscribe(additional_channel)
+                    additional_channel = (await redis_client.subscribe(message))[0]
+                    asyncio.create_task(ws_reader(additional_channel, ws))
+        except WebSocketDisconnect:
+            log.debug("Socket disconnected, no big deal")
+
+    redis = await aioredis.create_redis_pool(
+        address=f"redis://{settings.redis_host}:{settings.redis_port}",
+        db=settings.in_app_notif_cache_id,
+        password=settings.redis_password
+    )
+    consumer_task = consumer_handler(websocket, redis)
+    done, pending = await asyncio.wait(
+        [consumer_task, ], return_when=asyncio.FIRST_COMPLETED,
+    )
+    log.debug(f"Done task: {done}")
+    for task in pending:
+        log.warning(f"Canceling task: {task}")
+        task.cancel()
+    redis.close()
+    await redis.wait_closed()
+
+
 def create_app():
     app = FastAPI(title=settings.app_name)
     es = Elasticsearch(settings.es_endpoint)
@@ -227,3 +323,20 @@ def create_app():
             allow_headers=["*"],
         )
     return app
+
+
+def publish_message(message: Union[Dict[Any, Any], str], wallet: str, sync_redis: object = None):
+    """Publish a message to a user identified by wallet"""
+    if not sync_redis:
+        sync_redis = redis.Redis(
+            host=settings.redis_host,
+            port=settings.redis_port,
+            db=settings.in_app_notif_cache_id,
+            password=settings.redis_password,
+        )
+    if isinstance(message, Dict):
+        sync_redis.publish(wallet, json.dumps(message, default=str))
+    else:
+        sync_redis.publish(wallet, json.dumps({
+            "message": message
+            }))
