@@ -11,7 +11,8 @@ app_info = collect_info_for_apps()
 sns_client = app_info['session'].client(
   'sns', region_name=app_info.get('aws_region'))
 sqs_client = app_info['session'].client(
-  'sqs', region_name=app_info.get('aws_region'))
+  'sqs', region_name=app_info.get('aws_region'),
+  endpoint_url=f"https://sqs.{app_info.get('aws_region')}.amazonaws.com")
 
 
 def report_event(
